from pathlib import Path
from qnxmount.etfs.parser import Parser
from functools import lru_cache
from collections import defaultdict
import logging

LOGGER = logging.getLogger(__name__)


class ETFS:
    """This class implements the basic functionality to create a Kaitai io stream.

    Args:
        stream: Kaitaistream containing the qnx6 file system.
        pagesize (int): page size.

    Attributes:
        stream: Kaitaistream containing the qnx6 file system.
        parser (Parser): Parser class automatically generated from .ksy file.
        page_size (int): page size.
        current_pages: 
        ftable: File table of the file system.
        dir_tree: Directory tree of the file system.
        path_to_fid: Dictionary to map a file path to a fid.
    """
    def __init__(self, stream, pagesize=2048):
        self.stream = stream
        self.parser = Parser(self.stream)

        transaction_size = 16
        self.page_size = pagesize
        assert self.stream._io.size() % (pagesize + transaction_size) == 0, 'size of memory map invalid'
        num_pages = self.stream._io.size() // (pagesize + transaction_size)
        self.parser._m_pages = self.parser.Pages(pagesize, num_pages, self.parser._io, _root=self.parser)
        self.current_pages = self.get_current_file_pages()

        self.ftable = self.build_ftable()
        self.set_full_file_names()
        self.dir_tree = self.build_dir_tree()

        self.path_to_fid = {self.get_path(fid): fid for fid, entry in enumerate(self.ftable) if ((entry is not None) and (entry.full_name is not None))}

    def read_dir(self, path):
        return self.dir_tree[path]

    def build_dir_tree(self):
        tree = defaultdict(set)
        for fid, entry in enumerate(self.ftable):
            if entry is None or entry.full_name is None:
                continue
            path = self.get_path(fid)
            if fid != 0:
                tree[path.parent].add(fid)
        return tree

    @lru_cache(maxsize=128)
    def get_path(self, fid: int):
        if fid == 0:
            return Path('/')

        entry = self.ftable[fid]
        if self.ftable[entry.pfid] is None:
            parent_path = Path('/recovered_files')
        else:
            parent_path = self.get_path(entry.pfid)
        return parent_path / Path(entry.full_name)

    @lru_cache(maxsize=256)
    def read_file(self, fid):
        current_pages = self.current_pages[fid]
        data = b''
        for cluster in range(0 if not current_pages else max(current_pages) + 1):
            if cluster not in current_pages:
                LOGGER.warning(f"cluster {cluster} not found for {self.get_path(fid)} (fid {fid}). Replaced with 0xff.")
                data += b'\xff' * self.page_size
            else:
                data += current_pages[cluster].data.raw
        return data

    def set_full_file_names(self):
        for fid, entry in enumerate(self.ftable):
            if entry is None:
                continue
            if entry.is_extension or not entry.is_valid:
                entry._m_full_name = None

            elif entry.is_solo:
                entry._m_full_name = entry.body.name

            else:
                if self.ftable[entry.efid] is not None:
                    extension = self.ftable[entry.efid]
                    assert extension.is_extension
                    assert extension.pfid == fid
                    entry._m_full_name = entry.body.name + extension.body.name
                else:
                    entry._m_full_name = entry.body.name

    def get_current_file_pages(self):
        pages = defaultdict(lambda: defaultdict(lambda: defaultdict(None)))
        for page in self.parser.pages.pages:
            transaction = page.transaction
            if transaction.fid == 0xffff:
                continue
            pages[transaction.fid][transaction.cluster][transaction.sequence] = page

        for fid, clusters in pages.items():
            for cluster, sequences in clusters.items():
                # assume that for equal sequences the one with the highest offset is the most recent one
                # not sure if correct
                pages[fid][cluster] = sequences[max(sequences)]

        return pages

    def build_ftable(self):
        ftable = []
        ftable_pages = self.current_pages[self.parser.fid_ftable]
        for cluster in range(max(ftable_pages) + 1):
            if cluster not in ftable_pages:
                for _ in range(self.page_size // 64):
                    ftable.append(None)
                continue
            page = ftable_pages[cluster]
            for entry in page.data.as_ftable.entries:
                ftable.append(entry)
        return ftable

if __name__ == '__main__':
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument('image', type=Path, help='Path to EFS binary')
    parser.add_argument('-s', '--page_size', type=lambda x: int(x, 0), help='Size of pages (clusters)', default=2048)
    args = parser.parse_args()

    etfs = ETFS(args.image, 0, args.page_size)

    exit(0)





import errno
import logging
from pathlib import Path

from fuse import FUSE, FuseOSError, Operations

from qnxmount.etfs.interface import ETFS
from qnxmount.stream import Stream

LOGGER = logging.getLogger(__name__)


class FuseETFS(Operations):
    """Fuse implementation of the ETFS file system.

    Args:
        stream: Kaitaistream containing the qnx6 file system.
    """

    def __init__(self, stream, pagesize):
        self.etfs = ETFS(stream, pagesize=pagesize)

    def getattr(self, path, fh=None):
        """Get directory with stat information.

        Args:
            path (Path): Path.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"getattr({path})")
        path = Path(path)

        if path not in self.etfs.path_to_fid:
            raise FuseOSError(errno.ENOENT)

        fid = self.etfs.path_to_fid[path]
        entry = self.etfs.ftable[fid]
        return dict(
            st_size=entry.body.size,
            st_nlink=1,
            st_mode=entry.body.mode,
            st_ctime=entry.body.ctime,
            st_mtime=entry.body.mtime,
            st_uid=entry.body.uid,
            st_gid=entry.body.gid,
        )

    def readdir(self, path, fh):
        """Read content from directory.

        Args:
            path (Path): Path to directory.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"readdir({path})")
        path = Path(path)

        if path in self.etfs.dir_tree:
            for fid in self.etfs.dir_tree[path]:
                entry = self.etfs.ftable[fid]
                yield entry.full_name

    def read(self, path, size, offset, fh):
        """Read content from an object.

        Args:
            path (Path): Path to object.
            size (int): size of content to be read.
            offset (int): starting offset in the file.
            fh: file handle.

        Returns:
            bytes: Raw file content.
        """
        LOGGER.debug(f"read({path}, {size}, {offset})")
        path = Path(path)

        if not path in self.etfs.path_to_fid:
            raise FuseOSError(errno.ENOENT)

        fid = self.etfs.path_to_fid[path]
        file = self.etfs.read_file(fid)
        return file[offset : offset + size]

    def readlink(self, path):
        """Read content from a link given the path.

        Args:
            path (Path): Path to symlink.

        Returns:
            str: symlink target
        """
        LOGGER.debug(f"readlink({path})")
        path = Path(path)

        if not path in self.etfs.path_to_fid:
            raise FuseOSError(errno.ENOENT)

        fid = self.etfs.path_to_fid[path]
        return self.etfs.read_file(fid).decode("utf-8")


def mount(image, mount_point, offset, pagesize):
    with Stream(image, offset) as stream:
        etfs = FuseETFS(stream, pagesize)
        FUSE(etfs, str(mount_point), nothreads=True, foreground=True)



from qnxmount.qnx6.parser import Parser
from kaitaistruct import KaitaiStream, BytesIO
import stat
import crcmod
from functools import lru_cache


class QNX6FS:
    """This class implements the logic of the qnx6 file system and uses the
    structures in the class :class:`Parser`.

    Args:
        stream: Kaitaistream containing the qnx6 file system.

    Attributes:
        stream (KaitaiStream): Kaitai data stream of qnx6 file system.
        parser (Parser): Parser class automatically generated from .ksy file.
        blocksize (int): Block size.
        active_superblock (Parser.Superblock): Active superblock.
    """
    def __init__(self, stream):

        self.stream = stream
        self.parser = Parser(self.stream)
        self.check_superblock_crc()

        self.blocksize = self.parser.blocksize
        if self.parser.qnx6_bootblock.superblock0.serial >= self.parser.qnx6_bootblock.superblock1.serial:
            self.active_superblock = self.parser.qnx6_bootblock.superblock0
        else:
            self.active_superblock = self.parser.qnx6_bootblock.superblock1

    def get_dir_from_path(self, path):
        """Get directory content from its path.

        Args:
            path (Union[Path, PurePath]): Path to directory.

        Returns:
            Parser.Directory: Parsed directory.
        """
        inode_number = self.get_inode_number_from_path(path)
        return self.get_dir(inode_number)

    @lru_cache(maxsize=128)
    def get_inode_number_from_path(self, path):
        """Get the inode number from a path.

        Args:
            path (Union[Path, PurePath]): Path to object.

        Returns:
            int: Inode number.
        """
        if not path.name:
            return 1

        inode_number = self.get_inode_number_from_path(path.parent)
        directory = self.get_dir(inode_number)
        for entry in directory.entries:
            if entry.content.name == path.name:
                return entry.inode_number

        return None

    def get_longname(self, index):
        """Get longname from its index.

        Args:
            index (int): index in the longfile inode.

        Returns:
            Parser.Longname: Longname object.
        """
        longname_raw = self.read_file(self.active_superblock.longfile,
                                      offset=index * self.parser.blocksize, size=self.parser.blocksize)
        longname = self.parser.Longname(KaitaiStream(BytesIO(longname_raw)), _root=self.parser._root)
        return longname

    def get_dir(self, inode_number):
        """Get directory content from its inode number.

        Args:
            inode_number (int): Inode number of the directory.

        Returns:
            Parser.Directory: Parsed directory.
        """
        assert inode_number >= 1, 'Inode number smaller than 1 not allowed!'
        dir_raw = self.read_dir(inode_number)
        directory = self.parser.Directory(KaitaiStream(BytesIO(dir_raw)), _root=self.parser._root)
        directory.entries = directory.entries[:-1]

        valid_entries = []
        for idx, entry in enumerate(directory.entries):
            if entry.inode_number == 0:
                continue
            if not entry.content.name:
                longname = self.get_longname(entry.content.index)
                entry.content._m_name = longname.name
                entry.length = longname.length
            valid_entries.append(entry)

        directory.entries = valid_entries
        return directory

    def read_dir(self, inode_number):
        """Read raw directory content from its inode number.

        Args:
            inode_number (int): Inode number of the directory.

        Returns:
            bytes: Raw directory content.
        """
        assert inode_number >= 1, 'Inode number smaller than 1 not allowed!'
        dir_inode = self.get_inode(inode_number)
        assert stat.S_ISDIR(dir_inode.mode), 'Inode is not a directory!'
        dir_raw = self.read_file(dir_inode)

        return dir_raw

    def get_inode(self, inode_number):
        """Retrieve inode from its inode number.

        Args:
            inode_number (int): Inode number of the directory.

        Returns:
            Parser.Inode: Parsed inode.
        """
        # inode_number 1 is the root ("/") directory
        assert inode_number >= 1, 'Inode number smaller than 1 not allowed!'

        inode_raw = self.read_file(self.active_superblock.inodes,
                                   offset=(inode_number - 1)*self.parser.sizeof_inode, size=self.parser.sizeof_inode)

        return self.parser.Inode(KaitaiStream(BytesIO(inode_raw)), _root=self.parser._root)

    def read_file(self, inode, offset=0, size=None):
        """Read raw content from an inode.

        Args:
            inode (Parser.Inode): Inode number of the directory.
            offset (int): starting offset in file.
            size (int): size of content to be read.

        Returns:
            bytes: Raw inode content.
        """
        if size is None:
            size = inode.size - offset
        elif offset + size > inode.size:
            size = inode.size - offset
        # assert (offset <= inode.size and size <= inode.size - offset), 'Attempting to read beyond file size!'

        file_content = b''
        for o in range(offset, offset + size, self.blocksize):
            file_content += self.parse_block_pointer(inode, o).raw_body

        return file_content[offset % self.blocksize:offset % self.blocksize + size]

    def parse_block_pointer(self, inode, offset):
        """Get pointer to the data of the inode on the given offset.

        The inodes have space for 16 pointers to data blocks. When these file size is larger than
        16 blocks, the file system uses a layered method. The initial 16 pointers point to a block
        of pointers that point to data blocks (level 1). Level 0 is when the 16 pointers directly
        point to data blocks. The maximum level for an inode is 2.

        Args:
            inode (Parser.Inode): Inode object.
            offset: offset in the content of the inode.

        Returns:
            Parser.BlockPointer: Pointer to the data block at the given offset.
        """
        def bytes_in_unit(level): return self.blocksize * (self.blocksize // 4) ** (inode.level - level)
        idx = offset // bytes_in_unit(0)
        pointer = inode.pointers[idx]

        for lvl in range(inode.level):
            offset = offset % bytes_in_unit(lvl)
            idx = offset // bytes_in_unit(lvl + 1)
            pointer = pointer.body_as_pointer_list[idx]

        return pointer

    def check_superblock_crc(self):
        """Assert superblock integrity."""
        superblock0 = {"raw": self.parser.qnx6_bootblock.superblock0_raw,
                       "object": self.parser.qnx6_bootblock.superblock0}
        superblock1 = {"raw": self.parser.qnx6_bootblock.superblock1_raw,
                       "object": self.parser.qnx6_bootblock.superblock1}

        crc32_func = crcmod.mkCrcFun(0x104c11db7, initCrc=0, rev=False)

        for superblock in [superblock0, superblock1]:
            assert len(superblock["raw"]) == 512, "Incorrect superblock size!"
            assert superblock["object"].crc == crc32_func(superblock["raw"][8:]), "Incorrect superblock CRC!"




import logging
import subprocess
import sys
from getpass import getuser
from pathlib import Path
from socket import gethostname

FORMATTER = logging.Formatter("%(asctime)s — %(name)s — %(levelname)s — %(message)s")


def setup_logging(logger, loglevel=logging.INFO):
    logger.setLevel(loglevel)
    logger.addHandler(get_console_handler())
    # logger.propagate = False

    logger.info("Logging started")
    logger.info("Python version: %s", sys.version)
    logger.info("Script started by %s on %s", getuser(), gethostname())
    logger.info("Called with sys.argv:\n %s", " ".join(sys.argv))
    rev_hash = get_git_revision_hash(Path(sys.argv[0]).parent)
    if rev_hash:
        logger.info("Current commit hash: %s", rev_hash)
    else:
        logger.warning("This script is not in a git repository!")

    return logger


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_git_revision_hash(git_dir):
    try:
        output = subprocess.check_output(["git", "-C", str(git_dir), "rev-parse", "HEAD"], stderr=subprocess.DEVNULL)
        return output.decode("utf8").rstrip()
    except subprocess.CalledProcessError:
        pass
    return None



import logging
import sys
from argparse import ArgumentParser
from pathlib import Path

import qnxmount.efs.mount as efs
import qnxmount.etfs.mount as etfs
import qnxmount.qnx6.mount as qnx6
from qnxmount.logger import setup_logging

LOGGER = logging.getLogger("qnxmount")


def mount(args):
    LOGGER.info(f"Selected mounter type: {args.type}")
    LOGGER.info(f"Mounting image {args.image} on mount point {args.mount_point}")
    if args.type == "qnx6":
        qnx6.mount(args.image, args.mount_point, args.offset)
    elif args.type == "efs":
        efs.mount(args.image, args.mount_point)
    elif args.type == "etfs":
        etfs.mount(args.image, args.mount_point, args.offset, args.page_size)
    LOGGER.info(f"Unmounting image {args.image} from mount point {args.mount_point}")


if __name__ == "__main__":
    parent_parser = ArgumentParser(description="The parent parser", add_help=False)
    parent_parser.add_argument("image", type=Path, help="Path to image containing qnx file system")
    parent_parser.add_argument("mount_point", type=Path, help="Path to mount point")

    main_parser = ArgumentParser(prog="qnxmount")
    subparsers = main_parser.add_subparsers(title="file system types", required=True, dest="type")
    parser_qnx6 = subparsers.add_parser("qnx6", parents=[parent_parser], help="Parser for HDD/eMMC images")
    parser_qnx6.add_argument("-o", "--offset", type=lambda x: int(x, 0), help="Offset of qnx partition in image", default=0)
    parser_efs = subparsers.add_parser("efs", parents=[parent_parser], help="Parser for NOR flash images")
    parser_etfs = subparsers.add_parser("etfs", parents=[parent_parser], help="Parser for NAND flash images")
    parser_etfs.add_argument("-o", "--offset", type=lambda x: int(x, 0), help="Offset of qnx partition in image", default=0)
    parser_etfs.add_argument("-s", "--page_size", type=lambda x: int(x, 0), help="Size of pages (clusters)", default=2048)

    args = main_parser.parse_args()

    setup_logging(LOGGER)

    if not args.image.exists():
        LOGGER.info(f"Image file {args.image} not found, exiting.")
        sys.exit(-1)

    if not args.mount_point.exists():
        LOGGER.info(f"Mount point {args.mount_point} not found, exiting.")
        sys.exit(-2)

    sys.exit(mount(args))



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Parser(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        pass

    class Userdata(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def as_ftable(self):
            if hasattr(self, '_m_as_ftable'):
                return self._m_as_ftable

            _pos = self._io.pos()
            self._io.seek(0)
            self._m_as_ftable = Parser.Ftable(self._io, self, self._root)
            self._io.seek(_pos)
            return getattr(self, '_m_as_ftable', None)

        @property
        def raw(self):
            if hasattr(self, '_m_raw'):
                return self._m_raw

            _pos = self._io.pos()
            self._io.seek(0)
            self._m_raw = self._io.read_bytes_full()
            self._io.seek(_pos)
            return getattr(self, '_m_raw', None)


    class NoEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass


    class FtableEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.efid = self._io.read_u2le()
            self.pfid = self._io.read_u2le()
            _on = self.entry_type
            if _on == 0:
                self.body = Parser.FileEntry(self._io, self, self._root)
            elif _on == 1:
                self.body = Parser.ExtnameEntry(self._io, self, self._root)
            else:
                self.body = Parser.NoEntry(self._io, self, self._root)

        @property
        def has_no_parent(self):
            """These are deleted files or not-initialised entries."""
            if hasattr(self, '_m_has_no_parent'):
                return self._m_has_no_parent

            self._m_has_no_parent = self.pfid == 65535
            return getattr(self, '_m_has_no_parent', None)

        @property
        def is_extension(self):
            if hasattr(self, '_m_is_extension'):
                return self._m_is_extension

            self._m_is_extension = self.efid == 32768
            return getattr(self, '_m_is_extension', None)

        @property
        def full_name(self):
            if hasattr(self, '_m_full_name'):
                return self._m_full_name

            self._m_full_name = 0
            return getattr(self, '_m_full_name', None)

        @property
        def test(self):
            if hasattr(self, '_m_test'):
                return self._m_test

            self._m_test = (self.efid & 32768) == 32768
            return getattr(self, '_m_test', None)

        @property
        def is_solo(self):
            if hasattr(self, '_m_is_solo'):
                return self._m_is_solo

            self._m_is_solo = self.efid == 0
            return getattr(self, '_m_is_solo', None)

        @property
        def is_valid(self):
            if hasattr(self, '_m_is_valid'):
                return self._m_is_valid

            self._m_is_valid = self.efid != 65535
            return getattr(self, '_m_is_valid', None)

        @property
        def entry_type(self):
            """Dummy variable to switch on."""
            if hasattr(self, '_m_entry_type'):
                return self._m_entry_type

            self._m_entry_type = (int(self.is_extension) + (int(self.has_no_parent) * 2))
            return getattr(self, '_m_entry_type', None)


    class TransCode(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.code = self._io.read_u1()

        @property
        def deverr(self):
            if hasattr(self, '_m_deverr'):
                return self._m_deverr

            self._m_deverr = (self.code & 15) == 6
            return getattr(self, '_m_deverr', None)

        @property
        def dataerr(self):
            if hasattr(self, '_m_dataerr'):
                return self._m_dataerr

            self._m_dataerr = (self.code & 15) == 5
            return getattr(self, '_m_dataerr', None)

        @property
        def erased(self):
            if hasattr(self, '_m_erased'):
                return self._m_erased

            self._m_erased = (self.code & 15) == 2
            return getattr(self, '_m_erased', None)

        @property
        def ok(self):
            if hasattr(self, '_m_ok'):
                return self._m_ok

            self._m_ok = (self.code & 15) == 0
            return getattr(self, '_m_ok', None)

        @property
        def ecc(self):
            if hasattr(self, '_m_ecc'):
                return self._m_ecc

            self._m_ecc = (self.code & 15) == 1
            return getattr(self, '_m_ecc', None)

        @property
        def foxes(self):
            if hasattr(self, '_m_foxes'):
                return self._m_foxes

            self._m_foxes = (self.code & 15) == 3
            return getattr(self, '_m_foxes', None)

        @property
        def badblk(self):
            if hasattr(self, '_m_badblk'):
                return self._m_badblk

            self._m_badblk = (self.code & 15) == 7
            return getattr(self, '_m_badblk', None)


    class Page(KaitaiStruct):
        def __init__(self, pagesize, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.pagesize = pagesize
            self._read()

        def _read(self):
            self._raw_data = self._io.read_bytes(self.pagesize)
            _io__raw_data = KaitaiStream(BytesIO(self._raw_data))
            self.data = Parser.Userdata(_io__raw_data, self, self._root)
            self.transaction = Parser.Transaction(self._io, self, self._root)


    class Pages(KaitaiStruct):
        def __init__(self, pagesize, num_pages, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.pagesize = pagesize
            self.num_pages = num_pages
            self._read()

        def _read(self):
            self.pages = []
            for i in range(self.num_pages):
                self.pages.append(Parser.Page(self.pagesize, self._io, self, self._root))



    class Ftable(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.entries = []
            i = 0
            while not self._io.is_eof():
                self.entries.append(Parser.FtableEntry(self._io, self, self._root))
                i += 1



    class FileEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.mode = self._io.read_u4le()
            self.uid = self._io.read_u4le()
            self.gid = self._io.read_u4le()
            self.atime = self._io.read_u4le()
            self.mtime = self._io.read_u4le()
            self.ctime = self._io.read_u4le()
            self.size = self._io.read_u4le()
            self.name = (KaitaiStream.bytes_terminate(self._io.read_bytes(32), 0, False)).decode(u"utf-8")


    class Transaction(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.fid = self._io.read_u4le()
            self.cluster = self._io.read_u4le()
            self.nclusters = self._io.read_u2le()
            self.tacode = Parser.TransCode(self._io, self, self._root)
            self.dacode = Parser.TransCode(self._io, self, self._root)
            self.sequence = self._io.read_u4le()


    class ExtnameEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.name = (KaitaiStream.bytes_terminate(self._io.read_bytes(59), 0, False)).decode(u"utf-8")
            self.type = self._io.read_u1()

        @property
        def is_valid(self):
            if hasattr(self, '_m_is_valid'):
                return self._m_is_valid

            self._m_is_valid = self.type == 0
            return getattr(self, '_m_is_valid', None)


    @property
    def fid_firstfile(self):
        if hasattr(self, '_m_fid_firstfile'):
            return self._m_fid_firstfile

        self._m_fid_firstfile = 6
        return getattr(self, '_m_fid_firstfile', None)

    @property
    def pages(self):
        if hasattr(self, '_m_pages'):
            return self._m_pages

        _pos = self._io.pos()
        self._io.seek(0)
        self._m_pages = Parser.Pages(2048, 9, self._io, self, self._root)
        self._io.seek(_pos)
        return getattr(self, '_m_pages', None)

    @property
    def fid_badblks(self):
        if hasattr(self, '_m_fid_badblks'):
            return self._m_fid_badblks

        self._m_fid_badblks = 2
        return getattr(self, '_m_fid_badblks', None)

    @property
    def fid_ftable(self):
        if hasattr(self, '_m_fid_ftable'):
            return self._m_fid_ftable

        self._m_fid_ftable = 1
        return getattr(self, '_m_fid_ftable', None)

    @property
    def fid_root(self):
        if hasattr(self, '_m_fid_root'):
            return self._m_fid_root

        self._m_fid_root = 0
        return getattr(self, '_m_fid_root', None)

    @property
    def fid_counts(self):
        if hasattr(self, '_m_fid_counts'):
            return self._m_fid_counts

        self._m_fid_counts = 3
        return getattr(self, '_m_fid_counts', None)

    @property
    def fid_lostfound(self):
        if hasattr(self, '_m_fid_lostfound'):
            return self._m_fid_lostfound

        self._m_fid_lostfound = 4
        return getattr(self, '_m_fid_lostfound', None)

    @property
    def fid_reserved(self):
        if hasattr(self, '_m_fid_reserved'):
            return self._m_fid_reserved

        self._m_fid_reserved = 5
        return getattr(self, '_m_fid_reserved', None)





import errno
import logging
import stat
from pathlib import PurePath

from fuse import FUSE, FuseOSError, Operations

from qnxmount.qnx6.interface import QNX6FS
from qnxmount.stream import Stream

LOGGER = logging.getLogger(__name__)


class FuseQNX6(Operations):
    """Fuse implementation of the qnx6 file system.

    Args:
        stream: Kaitaistream containing the qnx6 file system.
    """

    def __init__(self, stream):
        self.qnx6fs = QNX6FS(stream)

    def getattr(self, path, fh=None):
        """Get directory with inode information.

        Args:
            path (PurePath): Path to inode.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"getattr({path=})")
        path = PurePath(path)
        inode_number = self.qnx6fs.get_inode_number_from_path(path)
        if inode_number is None:
            raise FuseOSError(errno.ENOENT)

        inode = self.qnx6fs.get_inode(inode_number)
        return dict(
            st_size=inode.size,
            st_nlink=1,
            st_mode=inode.mode,
            st_ctime=inode.ctime,
            st_mtime=inode.mtime,
            st_uid=inode.uid,
            st_gid=inode.gid,
        )

    def readdir(self, path, fh):
        """Read content from directory inode.

        Args:
            path (PurePath): Path to inode.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"readdir({path=})")
        path = PurePath(path)
        directory = self.qnx6fs.get_dir_from_path(path)
        for entry in directory.entries:
            yield entry.content.name

    def read(self, path, size, offset, fh):
        """Read content from an inode.

        Args:
            path (PurePath): Path to inode.
            size (int): size of content to be read.
            offset (int): starting offset in the file.
            fh: file handle.

        Returns:
            bytes: Raw inode content.
        """
        LOGGER.debug(f"read({path=}, {size=}, {offset=})")
        path = PurePath(path)
        inode_number = self.qnx6fs.get_inode_number_from_path(path)
        if not inode_number:
            raise FuseOSError(errno.ENOENT)

        inode = self.qnx6fs.get_inode(inode_number)
        if inode and stat.S_ISREG(inode.mode):
            return self.qnx6fs.read_file(inode, offset, size)
        raise FuseOSError(errno.EIO)

    def readlink(self, path):
        """Read content from a link given the path.

        Args:
            path (PurePath): Path to inode.

        Returns:
            str: symlink target
        """
        LOGGER.debug(f"readlink({path=})")
        path = PurePath(path)
        inode_number = self.qnx6fs.get_inode_number_from_path(path)
        if not inode_number:
            raise FuseOSError(errno.ENOENT)

        inode = self.qnx6fs.get_inode(inode_number)
        if inode and stat.S_ISLNK(inode.mode):
            link_raw = self.qnx6fs.read_file(inode)
            return link_raw.decode("utf8")
        raise FuseOSError(errno.ENOENT)


def mount(image, mount_point, offset):
    with Stream(image, offset) as stream:
        qnx6 = FuseQNX6(stream)
        FUSE(qnx6, str(mount_point), nothreads=True, foreground=True)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Parser(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        pass

    class Extent(KaitaiStruct):
        def __init__(self, i, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.i = i
            self._read()

        def _read(self):
            pass

        @property
        def text_offset(self):
            if hasattr(self, '_m_text_offset'):
                return self._m_text_offset

            self._m_text_offset = (((self.header.text_offset_hi << 16) + self.header.text_offset_lo) << self._parent._parent.align_pow2)
            return getattr(self, '_m_text_offset', None)

        @property
        def text(self):
            if hasattr(self, '_m_text'):
                return self._m_text

            _pos = self._io.pos()
            self._io.seek(self.text_offset)
            self._m_text = self._io.read_bytes(self.header.text_size)
            self._io.seek(_pos)
            return getattr(self, '_m_text', None)

        @property
        def as_dir_entry(self):
            if hasattr(self, '_m_as_dir_entry'):
                return self._m_as_dir_entry

            _pos = self._io.pos()
            self._io.seek(self.text_offset)
            self._raw__m_as_dir_entry = self._io.read_bytes(self.header.text_size)
            _io__raw__m_as_dir_entry = KaitaiStream(BytesIO(self._raw__m_as_dir_entry))
            self._m_as_dir_entry = Parser.DirEntry(_io__raw__m_as_dir_entry, self, self._root)
            self._io.seek(_pos)
            return getattr(self, '_m_as_dir_entry', None)

        @property
        def header(self):
            if hasattr(self, '_m_header'):
                return self._m_header

            _pos = self._io.pos()
            self._io.seek((self._root.unit_size - (32 * (self.i + 1))))
            self._raw__m_header = self._io.read_bytes(32)
            _io__raw__m_header = KaitaiStream(BytesIO(self._raw__m_header))
            self._m_header = Parser.ExtentHeader(_io__raw__m_header, self, self._root)
            self._io.seek(_pos)
            return getattr(self, '_m_header', None)

        @property
        def special_text(self):
            if hasattr(self, '_m_special_text'):
                return self._m_special_text

            _pos = self._io.pos()
            self._io.seek(self.text_offset)
            self._raw__m_special_text = self._io.read_bytes(self.header.text_size)
            _io__raw__m_special_text = KaitaiStream(BytesIO(self._raw__m_special_text))
            self._m_special_text = Parser.SpecialExtentP(self.i, _io__raw__m_special_text, self, self._root)
            self._io.seek(_pos)
            return getattr(self, '_m_special_text', None)


    class UnitInfo(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.struct_size = self._io.read_u2le()
            self.endian = self._io.read_bytes(1)
            self.pad = self._io.read_bytes(1)
            if not self.pad == b"\xFF":
                raise kaitaistruct.ValidationNotEqualError(b"\xFF", self.pad, self._io, u"/types/unit_info/seq/2")
            self.unit_pow2 = self._io.read_u2le()
            self.reserve = self._io.read_bytes(2)
            if not self.reserve == b"\xFF\xFF":
                raise kaitaistruct.ValidationNotEqualError(b"\xFF\xFF", self.reserve, self._io, u"/types/unit_info/seq/4")
            self.erase_count = self._io.read_u4le()
            self.boot = Parser.ExtentPtr(self._io, self, self._root)

        @property
        def unit_size(self):
            if hasattr(self, '_m_unit_size'):
                return self._m_unit_size

            self._m_unit_size = (1 << self.unit_pow2)
            return getattr(self, '_m_unit_size', None)


    class ExtentPtr(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.logi_unit = self._io.read_u2le()
            self.index = self._io.read_u2le()


    class SpecialExtentP(KaitaiStruct):
        def __init__(self, idx, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.idx = idx
            self._read()

        def _read(self):
            _on = self.idx
            if _on == 0:
                self.body = Parser.UnitInfo(self._io, self, self._root)
            elif _on == 4:
                self.body = Parser.DirEntry(self._io, self, self._root)
            elif _on == 1:
                self.body = Parser.UnitLogi(self._io, self, self._root)
            elif _on == 3:
                self.body = Parser.DirEntry(self._io, self, self._root)
            elif _on == 2:
                self.body = Parser.BootInfo(self._io, self, self._root)


    class Stat(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.struct_size = self._io.read_u2le()
            self.mode = self._io.read_u2le()
            self.uid = self._io.read_u4le()
            self.gid = self._io.read_u4le()
            self.mtime = self._io.read_u4le()
            self.ctime = self._io.read_u4le()


    class UnitLogi(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.struct_size = self._io.read_u2le()
            self.logi = self._io.read_u2le()
            self.age = self._io.read_u4le()
            self.pad = self._io.read_bytes((self.struct_size - 24))
            self.md5 = self._io.read_bytes(16)


    class DirEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.struct_size = self._io.read_u2le()
            self.moves = self._io.read_u1()
            self.namelen = self._io.read_u1()
            self.first = Parser.ExtentPtr(self._io, self, self._root)
            self.name = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
            self.name_pad = self._io.read_bytes((((self.namelen + 3) & 252) - self.namelen))
            self.stat = Parser.Stat(self._io, self, self._root)


    class ExtentHeader(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.status0 = Parser.HeaderStatus(self._io, self, self._root)
            self.status1 = self._io.read_u4le()
            self.status2 = self._io.read_u4le()
            self.ecc = self._io.read_bytes(6)
            self.reserve = self._io.read_bytes(1)
            if not self.reserve == b"\xFF":
                raise kaitaistruct.ValidationNotEqualError(b"\xFF", self.reserve, self._io, u"/types/extent_header/seq/4")
            self.text_offset_hi = self._io.read_u1()
            self.text_offset_lo = self._io.read_u2le()
            self.text_size = self._io.read_u2le()
            self.next = Parser.ExtentPtr(self._io, self, self._root)
            self.super = Parser.ExtentPtr(self._io, self, self._root)


    class HeaderStatus(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.no_write = self._io.read_bits_int_le(1) != 0
            self.no_next = self._io.read_bits_int_le(1) != 0
            self.no_super = self._io.read_bits_int_le(1) != 0
            self.no_split = self._io.read_bits_int_le(1) != 0
            self.condition = self._io.read_bits_int_le(3)
            self.ext_last = self._io.read_bits_int_le(1) != 0
            self.type = self._io.read_bits_int_le(2)
            self.basic = self._io.read_bits_int_le(1) != 0
            self.pad = self._io.read_bits_int_le(21)

        @property
        def is_file(self):
            if hasattr(self, '_m_is_file'):
                return self._m_is_file

            self._m_is_file = self.type == 3
            return getattr(self, '_m_is_file', None)

        @property
        def is_dir(self):
            if hasattr(self, '_m_is_dir'):
                return self._m_is_dir

            self._m_is_dir = self.type == 2
            return getattr(self, '_m_is_dir', None)

        @property
        def is_sys(self):
            if hasattr(self, '_m_is_sys'):
                return self._m_is_sys

            self._m_is_sys = self.type == 1
            return getattr(self, '_m_is_sys', None)


    class BootInfo(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.struct_size = self._io.read_u2le()
            self.rev_major = self._io.read_u1()
            self.rev_minor = self._io.read_u1()
            self.sig = self._io.read_bytes(8)
            if not self.sig == b"\x51\x53\x53\x4C\x5F\x46\x33\x53":
                raise kaitaistruct.ValidationNotEqualError(b"\x51\x53\x53\x4C\x5F\x46\x33\x53", self.sig, self._io, u"/types/boot_info/seq/3")
            self.unit_index = self._io.read_u2le()
            self.unit_total = self._io.read_u2le()
            self.unit_spare = self._io.read_u2le()
            self.align_pow2 = self._io.read_u2le()
            self.root = Parser.ExtentPtr(self._io, self, self._root)

        @property
        def is_valid(self):
            if hasattr(self, '_m_is_valid'):
                return self._m_is_valid

            self._m_is_valid =  ((self.struct_size == 24) and (self.rev_major == 3) and (self.rev_minor == 0)) 
            return getattr(self, '_m_is_valid', None)


    class Units(KaitaiStruct):
        def __init__(self, align_pow2, num_units, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.align_pow2 = align_pow2
            self.num_units = num_units
            self._read()

        def _read(self):
            self._raw_unit_list = []
            self.unit_list = []
            for i in range(self.num_units):
                self._raw_unit_list.append(self._io.read_bytes(self._root.unit_size))
                _io__raw_unit_list = KaitaiStream(BytesIO(self._raw_unit_list[i]))
                self.unit_list.append(Parser.Unit(_io__raw_unit_list, self, self._root))



    class Unit(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def raw(self):
            if hasattr(self, '_m_raw'):
                return self._m_raw

            _pos = self._io.pos()
            self._io.seek(0)
            self._m_raw = self._io.read_bytes(self._root.unit_size)
            self._io.seek(_pos)
            return getattr(self, '_m_raw', None)

        @property
        def extents(self):
            if hasattr(self, '_m_extents'):
                return self._m_extents

            self._m_extents = []
            i = 0
            while True:
                _ = Parser.Extent(i, self._io, self, self._root)
                self._m_extents.append(_)
                if _.header.status0.ext_last:
                    break
                i += 1
            return getattr(self, '_m_extents', None)

        @property
        def info(self):
            if hasattr(self, '_m_info'):
                return self._m_info

            self._m_info = self.extents[0].special_text.body
            return getattr(self, '_m_info', None)

        @property
        def logi(self):
            if hasattr(self, '_m_logi'):
                return self._m_logi

            self._m_logi = self.extents[1].special_text.body
            return getattr(self, '_m_logi', None)


    @property
    def unit_size(self):
        if hasattr(self, '_m_unit_size'):
            return self._m_unit_size

        self._m_unit_size = self.unit_shortcut.unit_size
        return getattr(self, '_m_unit_size', None)

    @property
    def unit_shortcut(self):
        """Needed for bootstrapping the unit size value."""
        if hasattr(self, '_m_unit_shortcut'):
            return self._m_unit_shortcut

        _pos = self._io.pos()
        self._io.seek(0)
        self._m_unit_shortcut = Parser.UnitInfo(self._io, self, self._root)
        self._io.seek(_pos)
        return getattr(self, '_m_unit_shortcut', None)

    @property
    def units(self):
        """Dummy values to satisfy the compiler, we will manually instantiate it anyway."""
        if hasattr(self, '_m_units'):
            return self._m_units

        self._m_units = Parser.Units(1, 1, self._io, self, self._root)
        return getattr(self, '_m_units', None)





import errno
import logging
import os
from collections import defaultdict
from pathlib import Path
from typing import Dict

from fuse import FUSE, FuseOSError, Operations

from qnxmount.efs.interface import EFS, scan_partitions

LOGGER = logging.getLogger(__name__)


class FuseEFS(Operations):
    """Fuse implementation of the EFS file system.

    Args:
        partitions (Dict[Path, EFS]): Dictionary of efs partitions
        fuse_mount_point (Path): Mount point of FUSE on the host OS.
        mount_point_dir_entries (Dict[Path, set]): Dictionary for paths contained in the
            mount point path of the partitions.
    """

    def __init__(self, partitions: Dict[Path, EFS], fuse_mount_point: Path):
        self.partitions = partitions
        self.fuse_mount_point = fuse_mount_point
        self.mount_point_dir_entries = self.build_dir_entry_tree()

    def build_dir_entry_tree(self):
        """Build a default tree for the dummy directory entries on the hard coded mount path to a partition.

        Returns:
            Dict[Path, set]: Dictionary with directory entries.
        """
        tree = defaultdict(set)
        for mount_point in self.partitions:
            path = mount_point
            while path.name:
                tree[path.parent].add(path.name)
                path = path.parent
        return tree

    def get_partition_mount_point(self, path):
        """Get mount point of the file system partition from a given path.

        Args:
            path (Path): Path in the file system.

        Returns:
            Path: Mount point of the partition.
        """
        for mount_point in self.partitions:
            if mount_point in path.parents or path == mount_point:
                return mount_point
        return None

    def getattr(self, path, fh=None):
        """Get directory with stat information.

        Args:
            path (Path): Path.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"getattr({path})")
        path = Path(path)

        """
        Check if path is a valid prefix of a hard coded mount point.
        If so, return dummy attributes derived from the hosts mount point.
        """
        if path in self.mount_point_dir_entries:
            st = os.lstat(self.fuse_mount_point.parent)
            return dict(
                (key, getattr(st, key))
                for key in ("st_atime", "st_ctime", "st_gid", "st_mode", "st_mtime", "st_nlink", "st_size", "st_uid")
            )

        partition_mount_point = self.get_partition_mount_point(path)
        if partition_mount_point is not None:
            path = path.relative_to(partition_mount_point)
            partition = self.partitions[partition_mount_point]
            dir_entry = partition.get_dir_entry_from_path(path)
            if dir_entry:
                return dict(
                    st_size=len(partition.read_file(dir_entry)),
                    st_nlink=1,
                    st_mode=dir_entry.stat.mode,
                    st_ctime=dir_entry.stat.ctime,
                    st_mtime=dir_entry.stat.mtime,
                    st_uid=dir_entry.stat.uid,
                    st_gid=dir_entry.stat.gid,
                )

        raise FuseOSError(errno.ENOENT)

    def readdir(self, path, fh):
        """Read content from directory.

        Args:
            path (Path): Path to directory.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"readdir({path})")
        path = Path(path)

        """
        Check if path is a valid prefix of a hard coded mount point.
        If so, return the correct directory entries.
        """
        if path in self.mount_point_dir_entries:
            for entry in self.mount_point_dir_entries[path]:
                yield entry

        else:
            partition_mount_point = self.get_partition_mount_point(path)
            if partition_mount_point is not None:
                path = path.relative_to(partition_mount_point)
                partition = self.partitions[partition_mount_point]
                dir_entry = partition.get_dir_entry_from_path(path)
                for entry in partition.read_dir(dir_entry):
                    yield entry.name

    def read(self, path, size, offset, fh):
        """Read content from an object.

        Args:
            path (Path): Path to object.
            size (int): size of content to be read.
            offset (int): starting offset in the file.
            fh: file handle.

        Returns:
            bytes: Raw file content.
        """
        LOGGER.debug(f"read({path}, {size}, {offset})")
        path = Path(path)
        mountpoint = self.get_partition_mount_point(path)
        if mountpoint is None:
            return b""
        path = path.relative_to(mountpoint)
        partition = self.partitions[mountpoint]
        dir_entry = partition.get_dir_entry_from_path(path)
        return partition.read_file(dir_entry)[offset : offset + size]

    def readlink(self, path):
        """Read content from a link given the path.

        Args:
            path (Path): Path to symlink.

        Returns:
            str: symlink target
        """
        LOGGER.debug(f"readlink({path})")
        path = Path(path)
        mountpoint = self.get_partition_mount_point(path)
        if mountpoint is None:
            raise FuseOSError(errno.ENOENT)
        path = path.relative_to(mountpoint)
        partition = self.partitions[mountpoint]
        dir_entry = partition.get_dir_entry_from_path(path)
        return partition.read_file(dir_entry).decode("utf-8")


def initialise_fuse_efs(image, mount_point):
    """Initialise the FuseEFS object.

    Args:
        image (Path): Path to the image.
        mount_point (Path): Path to the mount point on the host system.

    Returns:
        FuseEFS: FuseEFS object.
    """
    partitions = dict()
    for partition in scan_partitions(image):
        mount_path = Path(partition.root.name)
        if not mount_path.is_absolute():
            mount_path = Path("/") / mount_path
        partitions[mount_path] = partition

    return FuseEFS(partitions, mount_point)


def mount(image, mount_point):
    efs = initialise_fuse_efs(image, mount_point)
    FUSE(efs, str(mount_point), nothreads=True, foreground=True)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Parser(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.qnx6_bootblock = Parser.Bootblock(self._io, self, self._root)

    class DirEntryLongname(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.zeros = self._io.read_bytes(3)
            self.index = self._io.read_u4le()
            self.checksum = self._io.read_u4le()
            self.more_zeros = self._io.read_bytes(16)

        @property
        def name(self):
            if hasattr(self, '_m_name'):
                return self._m_name

            self._m_name = 0
            return getattr(self, '_m_name', None)


    class Shortname(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.name = (self._io.read_bytes_term(0, False, True, False)).decode(u"UTF-8")


    class DirEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.inode_number = self._io.read_u4le()
            self.length = self._io.read_u1()
            _on = self.length
            if _on == 255:
                self._raw_content = self._io.read_bytes(27)
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = Parser.DirEntryLongname(_io__raw_content, self, self._root)
            else:
                self._raw_content = self._io.read_bytes(27)
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = Parser.Shortname(_io__raw_content, self, self._root)


    class Superblock(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.magic = self._io.read_bytes(4)
            if not self.magic == b"\x22\x11\x19\x68":
                raise kaitaistruct.ValidationNotEqualError(b"\x22\x11\x19\x68", self.magic, self._io, u"/types/superblock/seq/0")
            self.crc = self._io.read_u4le()
            self.serial = self._io.read_u8le()
            self.ctime = self._io.read_u4le()
            self.atime = self._io.read_u4le()
            self.flag = self._io.read_u4le()
            self.version1 = self._io.read_u2le()
            self.version2 = self._io.read_u2le()
            self.volumeid = self._io.read_bytes(16)
            self.blocksize = self._io.read_u4le()
            self.num_inodes = self._io.read_u4le()
            self.free_inodes = self._io.read_u4le()
            self.num_blocks = self._io.read_u4le()
            self.free_blocks = self._io.read_u4le()
            self.allocgroup = self._io.read_u4le()
            self.inodes = Parser.Rootnode(self._io, self, self._root)
            self.bitmap = Parser.Rootnode(self._io, self, self._root)
            self.longfile = Parser.Rootnode(self._io, self, self._root)
            self.iclaim = Parser.Rootnode(self._io, self, self._root)
            self.iextra = Parser.Rootnode(self._io, self, self._root)
            self.migrate_blocks = self._io.read_u4le()
            self.scrub_block = self._io.read_u4le()
            self.spare = self._io.read_bytes(32)


    class Inode(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.size = self._io.read_u8le()
            self.uid = self._io.read_u4le()
            self.gid = self._io.read_u4le()
            self.ftime = self._io.read_u4le()
            self.mtime = self._io.read_u4le()
            self.atime = self._io.read_u4le()
            self.ctime = self._io.read_u4le()
            self.mode = self._io.read_u2le()
            self.ext_mode = self._io.read_u2le()
            self.pointers = []
            for i in range(16):
                self.pointers.append(Parser.BlockPointer(self._io, self, self._root))

            self.level = self._io.read_u1()
            self.status = self._io.read_bytes(1)
            self.unknown = self._io.read_bytes(2)
            self.zeros = []
            for i in range(6):
                self.zeros.append(self._io.read_u4le())



    class Rootnode(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.size = self._io.read_u8le()
            self.pointers = []
            for i in range(16):
                self.pointers.append(Parser.BlockPointer(self._io, self, self._root))

            self.level = self._io.read_u1()
            self.mode = self._io.read_bytes(1)
            self.spare = self._io.read_bytes(6)


    class BlockPointer(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.ptr = self._io.read_u4le()

        @property
        def raw_body(self):
            if hasattr(self, '_m_raw_body'):
                return self._m_raw_body

            io = self._root._io
            _pos = io.pos()
            io.seek((self._root.data_start + (self.ptr * self._root.blocksize)))
            self._m_raw_body = io.read_bytes(self._root.blocksize)
            io.seek(_pos)
            return getattr(self, '_m_raw_body', None)

        @property
        def body_as_pointer_list(self):
            if hasattr(self, '_m_body_as_pointer_list'):
                return self._m_body_as_pointer_list

            io = self._root._io
            _pos = io.pos()
            io.seek((self._root.data_start + (self.ptr * self._root.blocksize)))
            self._m_body_as_pointer_list = []
            for i in range(self._root.blocksize // 4):
                self._m_body_as_pointer_list.append(Parser.BlockPointer(io, self, self._root))

            io.seek(_pos)
            return getattr(self, '_m_body_as_pointer_list', None)


    class Bootblock(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.magic = self._io.read_bytes(4)
            if not self.magic == b"\xEB\x10\x90\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\xEB\x10\x90\x00", self.magic, self._io, u"/types/bootblock/seq/0")
            self.offset_qnx6fs = self._io.read_u4le()
            self.sblk0 = self._io.read_u4le()
            self.sblk1 = self._io.read_u4le()

        @property
        def superblock0(self):
            if hasattr(self, '_m_superblock0'):
                return self._m_superblock0

            io = self._root._io
            _pos = io.pos()
            io.seek((self.sblk0 * 512))
            self._m_superblock0 = Parser.Superblock(io, self, self._root)
            io.seek(_pos)
            return getattr(self, '_m_superblock0', None)

        @property
        def superblock1(self):
            if hasattr(self, '_m_superblock1'):
                return self._m_superblock1

            io = self._root._io
            _pos = io.pos()
            io.seek((self.sblk1 * 512))
            self._m_superblock1 = Parser.Superblock(io, self, self._root)
            io.seek(_pos)
            return getattr(self, '_m_superblock1', None)

        @property
        def superblock0_raw(self):
            if hasattr(self, '_m_superblock0_raw'):
                return self._m_superblock0_raw

            io = self._root._io
            _pos = io.pos()
            io.seek((self.sblk0 * 512))
            self._m_superblock0_raw = io.read_bytes(512)
            io.seek(_pos)
            return getattr(self, '_m_superblock0_raw', None)

        @property
        def superblock1_raw(self):
            if hasattr(self, '_m_superblock1_raw'):
                return self._m_superblock1_raw

            io = self._root._io
            _pos = io.pos()
            io.seek((self.sblk1 * 512))
            self._m_superblock1_raw = io.read_bytes(512)
            io.seek(_pos)
            return getattr(self, '_m_superblock1_raw', None)


    class Longname(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.length = self._io.read_u2le()
            self.name = (self._io.read_bytes(self.length)).decode(u"UTF-8")


    class Directory(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.entries = []
            i = 0
            while True:
                _ = Parser.DirEntry(self._io, self, self._root)
                self.entries.append(_)
                if  ((_.inode_number == 0) and (_.length == 0)) :
                    break
                i += 1


    @property
    def sizeof_inode(self):
        if hasattr(self, '_m_sizeof_inode'):
            return self._m_sizeof_inode

        self._m_sizeof_inode = 128
        return getattr(self, '_m_sizeof_inode', None)

    @property
    def blocksize(self):
        if hasattr(self, '_m_blocksize'):
            return self._m_blocksize

        self._m_blocksize = self.qnx6_bootblock.superblock0.blocksize
        return getattr(self, '_m_blocksize', None)

    @property
    def abs_data_start_padding(self):
        """Kaitai does not support abs()."""
        if hasattr(self, '_m_abs_data_start_padding'):
            return self._m_abs_data_start_padding

        self._m_abs_data_start_padding = ((self.blocksize - 12288) if self.blocksize > 12288 else (12288 - self.blocksize))
        return getattr(self, '_m_abs_data_start_padding', None)

    @property
    def data_start(self):
        """https://github.com/RunZeJustin/qnx660/blob/47c4158e3993d7536170b649e6c1e09552318fb4/target/qnx6/usr/include/sys/fs_qnx6.h."""
        if hasattr(self, '_m_data_start'):
            return self._m_data_start

        self._m_data_start = (12288 + (0 if self.blocksize <= 4096 else self.abs_data_start_padding))
        return getattr(self, '_m_data_start', None)





from kaitaistruct import KaitaiStream
import mmap

class Stream:
    def __init__(self, path, offset=0):
        self.path = path
        self.offset = offset

    def __enter__(self):
        self.f = open(self.path, 'rb')
        self.mm = mmap.mmap(self.f.fileno(), length=0, access=mmap.ACCESS_READ, offset=self.offset)
        self.stream = KaitaiStream(self.mm)
        return self.stream

    def __exit__(self, type, value, traceback):
        self.stream.close()
        self.mm.close()
        self.f.close()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qnxmount', 'qnxmount.efs', 'qnxmount.etfs', 'qnxmount.qnx6']

package_data = \
{'': ['*']}

install_requires = \
['crcmod>=1.7,<2.0', 'fusepy>=3.0.1,<4.0.0', 'kaitaistruct>=0.10,<0.11']

setup_kwargs = {
    'name': 'qnxmount',
    'version': '0.1.3',
    'description': 'read only mounters for qnx filesystems',
    'long_description': '# QNX Filesystems Mounter\n\n## Project Discription\n\nThis project contains code to parse and mount (read only) QNX filesystems in non-standard images (HDD / SSD / eMMC).\n\nExisting tools were not able to handle the exotic configurations of some of the filesystems that we encountered in vehicle forensics, for instance on blocksizes greater than 4K on qnx6 filesystems, or non-standard allignment on qnx efs filesystems.\n\nThe description of the binary data structure of these filesystems is done with [kaitai](https://kaitai.io/) and this description can be found in the `.ksy` files in the folders for each respective qnx filesystem ([qnx6](qnxmount/qnx6/parser.ksy), [etfs](qnxmount/etfs/parser.ksy), and [efs](qnxmount/efs/parser.ksy)). With Kaitai, a Python based parser was generated. Mounting with these parsers is based on fuse.\n\nThis project is only tested on Linux machines. \n\n\n## Getting started\n\nSet up your Python virtual environment and activate the environment:\n```commandline\npython3 -m venv venv\nsource ./venv/bin/activate\n```\nInstall qnxmount and fuse in the virtual environment:\n```commandline\npip install qnxmount\nsudo apt install fuse\n```\n\n<!-- Or clone this repo and install.\n```commandline\npip install .\n``` -->\n\n\n## Usage\n\nGeneral use of the module is as follows:\n```shell\npython3 -m qnxmount {fs_type} [options] /image /mountpoint\n```\nwhere `fs_type` is the filesystem type (qnx6, etfs, or efs) and options are the options for that filesystem type.\n\nThe options are different for each filesystem type. An overview is given below. For more information use the help option. \n```shell\npython3 -m qnxmount qnx6 [-o OFFSET] /image /mountpoint\npython3 -m qnxmount etfs [-o OFFSET] [-s PAGE_SIZE] /image /mountpoint\npython3 -m efs /image /mountpoint\n```\n\nNote that the offset and page size can be entered in decimal, octal, binary, or hexadecimal format. For example, we can mount an image with a qnx6 filesystem at offset 0x1000 with:\n```shell\npython3 -m qnxmount qnx6 -o 0x1000 /image /mountpoint \n```\nUsing the option `-o 4096` would give the same result.\n\nIf mounting succeeds you will see the log message `"Mounting image /image on mount point /mountpoint"` appear and the process will hang. Navigate to the given mount point with another terminal session or a file browser to access the file system.\n\nUnmounting can be done from the terminal with:\n```shell\nsudo umount /mountpoint\n```\nThe logs will show show that the image was successfully unmounted and qnxmount will exit.\n\n## Contributing and Testing\n\nIf you want develop the tool and run tests, first fork the repository. Contributions can be submitted as a merge request. \n\nTo get started clone the forked repository and create a virtual environment. Install the test dependencies and fuse into the environment.\n```commandline\npip install .[test]\nsudo apt install fuse\n```\n\nThe folder **tests** contains functional tests to test the different parsers.\nTo run these tests you need a file system image and an accompanying tar archive.\nThe tests run are functional tests that check whether the parsed data from the test image is equal to the data stored in the archive.\nDefault test_images are located in the folders **test_data**.\nIf you want to test your own image replace the files **test_image.bin** and **test_image.tar.gz** with your own.\n\nA test image can be created by running the script `make_test_fs.sh` inside a QNX Virtual Machine.\nUpdate the script with the (edge) cases you want to check and run the command below.\nThis should create an _image.bin_ and _image.tar.gz_ into the specified directory.\nThese can be used as test files.\n```shell\nmake_test_fs.sh /path/to/output/directory\n```\n\nTo run the tests in this repo navigate to the main directory of the repo and run:\n```shell\npytest\n```\n\n[//]: # (Usually, tests can be run by directly calling `pytest tests --image ... --tar ...`, however this method fails here.)\n[//]: # (The reason is that the tests are located in a separate subfolder from the **qnx6_file_system.py**. )\n[//]: # (The qnx6_file_system module cannot be imported because it is not located in the tests directory.)\n[//]: # (When python3 is called it adds \'.\' to the PATH and since the qnx6_file_system module is located in the working directory they can be found.)\n',
    'author': 'Francis Hoogendijk',
    'author_email': 'f.hoogendijk@nfi.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/NetherlandsForensicInstitute/qnxmount',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from qnxmount.efs.parser import Parser
from kaitaistruct import KaitaiStream
from functools import lru_cache
import mmap
import re
import logging

LOGGER = logging.getLogger(__name__)


class KaitaiIO:
    """This class implements the basic functionality to create a Kaitai io stream.

    Args:
        path (Path): Path to the image.
        offset (int): Offset in the image to start the Kaitai stream at.

    Attributes:
        path (Path): Path to the image.
        f: File handle.
        mm (mmap): Memory map of the file.
        stream (KaitaiStream): Kaitai data stream of memory-mapped file.
    """
    def __init__(self, path, offset=0):
        self.path = path
        self.f = open(self.path, 'rb')
        self.mm = mmap.mmap(self.f.fileno(), length=0, offset=offset, access=mmap.ACCESS_READ)
        self.stream = KaitaiStream(self.mm)

    def __del__(self):
        self.stream.close()
        self.mm.close()
        self.f.close()


class EFS(KaitaiIO):
    """This class implements the logic of the EFS file system and uses the
    structures in the class :class:`Parser`.

    Args:
        path (Path): Path to image.
        boot: Bootinfo object of the EFS.
        offset: Start of first unit of the partition in image.

    Attributes:
        boot (Parser.BootInfo): Boot info.
        parser (Parser): Kaitai generated parser object.
        logi_map (Dict[int, Parser.Unit]: Logic mapping of units.
        root (Parser.DirEntry): Root of partition.
    """
    def __init__(self, path: Path, boot: Parser.BootInfo, offset: int = 0):
        super().__init__(path, offset)
        self.boot = boot
        self.parser = Parser(self.stream)
        self.parser._m_units = self.parser.Units(boot.align_pow2, boot.unit_total, self.parser._io, _root=self.parser)
        self.logi_map = self.get_logi_map()
        root_ext = self.get_ext_ptr(self.boot.root)
        self.root = root_ext.as_dir_entry

    def get_logi_map(self):
        """Create a mapping of units in logical order.

        Returns:
            Dict[int, EfsParset.Unit]: Dictionary with logical mapping.
        """
        logi_map = {unit.logi.logi: unit for unit in self.parser.units.unit_list if not is_spare(unit)}
        assert list(sorted(logi_map.keys())) == list(range(1, self.boot.unit_total - self.boot.unit_spare + 1)), 'Not all units are accounted for!'
        return logi_map

    def get_ext_ptr(self, ptr: Parser.ExtentPtr) -> Parser.Extent:
        """Get extent from extent pointer.

        Args:
            ptr: Extent pointer

        Returns:
            Parser.Extent: Extent
        """
        extent = self.logi_map[ptr.logi_unit].extents[ptr.index]
        while not extent.header.status0.no_super:
            ptr = extent.header.super
            extent = self.get_ext_ptr(ptr)
        return extent

    def get_extents(self, dir_entry: Parser.DirEntry):
        """Get all extents from a directory entry.

        Args:
            dir_entry (Parser.DirEntry): Directory entry.

        Yields:
            Parser.Extent: All relevant extents in the directory entry.
        """
        extent = self.get_ext_ptr(dir_entry.first)
        yield extent

        while not extent.header.status0.no_next:
            extent = self.get_ext_ptr(extent.header.next)
            yield extent

    def read_dir(self, dir_entry: Parser.DirEntry):
        """Read a directory.

        Args:
            dir_entry (Parser.DirEntry): Directory entry.

        Yields:
            Parser.DirEntry: Extent parsed as a directory entry.
        """
        for extent in self.get_extents(dir_entry):
            if not extent.header.text_size:
                break
            yield extent.as_dir_entry

    def stat(self, dir_entry: Parser.DirEntry):
        """Return stat info of a directory entry.

        Args:
            dir_entry (Parser.DirEntry): Directory entry.

        Returns:
            int: Mode.
        """
        mode = dir_entry.stat.mode
        return mode

    def read_file(self, dir_entry: Parser.DirEntry):
        """Read contents of a file.

        Args:
            dir_entry (Parser.DirEntry): Directory entry.

        Returns:
            bytes: Full file content.
        """
        return b''.join(extent.text for extent in self.get_extents(dir_entry))

    @lru_cache(maxsize=128)
    def get_dir_entry_from_path(self, path: Path) -> Parser.DirEntry:
        """Get dir entry from path.

        Args:
            path (Path): path.

        Returns:
            Parser.DirEntry: Directory entry.
        """
        if not path.name:
            return self.root
        parent_entry = self.get_dir_entry_from_path(path.parent)
        for entry in self.read_dir(parent_entry):
            if entry.name == path.name:
                return entry


def scan_partitions(path):
    """Scans a flash image for valid EFS partitions based on the boot_info.

    Args:
        path(Path): Path to the image.

    Yields:
        EFS: An initialized EFS parser object for each valid partition found.
    """
    kaitai_io = KaitaiIO(path)
    sig_pattern = re.compile(b'QSSL_F3S', re.MULTILINE)
    for hit in re.finditer(sig_pattern, kaitai_io.mm):
        stream = kaitai_io.stream
        stream.seek(hit.start() - 4)
        boot = Parser.BootInfo(stream)
        if not boot.is_valid:
            continue

        boot_unit_start = hit.start() & 0xfffff000
        stream.seek(boot_unit_start)
        unit_info = Parser.UnitInfo(stream)

        # TODO maybe add stricter checking if partition is valid
        start_offset = boot_unit_start - boot.unit_index * unit_info.unit_size
        LOGGER.info(f'Valid partition found at {hex(start_offset)} (with boot entry located at {hex(hit.start())})')
        yield EFS(kaitai_io.path, boot, offset=start_offset)


def is_spare(unit: Parser.Unit):
    """Check if EFS unit is a spare unit.

    Args:
        unit (Parser.Unit): EFS unit.

    Returns:
        bool: Boolean answer.
    """
    if ((len(unit.extents) == 1) or
            (len(unit.extents) == 2 and unit.extents[1].header.status1 == 0xffffffff)):
        return True
    return False
